# -*- coding: utf-8  -*-
#
# Copyright (C) 2009-2012 Ben Kurtovic <ben.kurtovic@verizon.net>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from os import expanduser
import re
from threading import RLock
from time import sleep, time

import oursql

from earwigbot import exceptions
from earwigbot.tasks import Task

class DRNClerkBot(Task):
    """A task to clerk for [[WP:DRN]]."""
    name = "drn_clerkbot"
    number = 19

    # Case status:
    STATUS_UNKNOWN = 0
    STATUS_NEW = 1
    STATUS_OPEN = 2
    STATUS_STALE = 3
    STATUS_NEEDASSIST = 4
    STATUS_REVIEW = 5
    STATUS_RESOLVED = 6
    STATUS_CLOSED = 7
    STATUS_ARCHIVE = 8

    def setup(self):
        """Hook called immediately after the task is loaded."""
        cfg = self.config.tasks.get(self.name, {})

        # Set some wiki-related attributes:
        self.title = cfg.get("title", "Wikipedia:Dispute resolution noticeboard")
        self.talk = cfg.get("talk", "Wikipedia talk:Dispute resolution noticeboard")
        self.volunteer_title = cfg.get("volunteers", "Wikipedia:Dispute resolution noticeboard/Volunteering")
        default_summary = "Updating $3 cases for the [[WP:DRN|dispute resolution noticeboard]]."
        self.summary = self.make_summary(cfg.get("summary", default_summary))

        # Templates used:
        templates = cfg.get("templates", {})
        self.tl_status = templates.get("status", "DR case status")
        self.tl_notify_party = templates.get("notifyParty", "DRN-notice")
        self.tl_notify_stale = templates.get("notifyStale", "DRN stale notice")
        self.tl_archive_top = templates.get("archiveTop", "DRN archive top")
        self.tl_archive_bottom = templates.get("archiveBottom",
                                               "DRN archive bottom")

        # Connection data for our SQL database:
        kwargs = cfg.get("sql", {})
        kwargs["read_default_file"] = expanduser("~/.my.cnf")
        self.conn_data = kwargs
        self.db_access_lock = RLock()

    def run(self, **kwargs):
        """Entry point for a task event."""
        if not self.db_access_lock.acquire(False):  # Non-blocking
            self.logger.info("A job is already ongoing; aborting")
            return
        action = kwargs.get("action", "all")
        try:
            start = time()
            conn = oursql.connect(**self.conn_data)
            site = self.bot.wiki.get_site()
            if action in ["all", "update_volunteers"]:
                self.update_volunteers(conn, site)
            if action in ["all", "clerk"]:
                log = u"Starting update to [[{0}]]".format(self.title)
                self.logger.info(log)
                cases = self.read_database(conn)
                page = site.get_page(self.title)
                text = page.get()
                self.read_page(conn, cases, text)
                noticies = self.clerk(conn, cases)
                self.save(page, cases, kwargs)
                self.send_notices(site, notices)
        finally:
            self.db_access_lock.release()

    def update_volunteers(self, conn, site):
        """Updates and stores the list of dispute resolution volunteers."""
        log = u"Updating volunteer list from [[{0}]]"
        self.logger.info(log.format(self.volunteer_title))
        page = site.get_page(self.volunteer_title)
        try:
            text = page.get()
        except exceptions.PageNotFoundError:
            text = ""
        marker = "<!-- please don't remove this comment (used by EarwigBot) -->"
        if marker not in text:
            log = u"The marker ({0}) wasn't found in the volunteer list at [[{1}]]!"
            self.logger.error(log.format(marker, page.title))
        text = text.split(marker)[1]
        additions = set()
        for line in text.splitlines():
            user = re.search("\# \{\{User\|(.*?)\}\}", line)
            if user:
                additions.add((user.group(1)))

        removals = set()
        query1 = "SELECT volunteer_username FROM volunteer"
        query2 = "DELETE FROM volunteer WHERE volunteer_username = ?"
        query3 = "INSERT INTO volunteer VALUES (?)"
        with conn.cursor() as cursor:
            cursor.execute(query1)
            for row in cursor:
                if row in additions:
                    additions.remove(row)
                else:
                    removals.add(row)
            if removals:
                cursor.executemany(query2, removals)
            if additions:
                cursor.executemany(query3, additions)

    def read_database(self, conn):
        """Return a list of _Cases from the database."""
        cases = []
        query = "SELECT case_id, case_title, case_status FROM case"
        with conn.cursor() as cursor:
            cursor.execute(query)
            for id_, name, status in cursor:
                case = _Case(id_, title, status)
                cases.append(case)
        return cases

    def read_page(self, conn, cases, text):
        """Read the noticeboard content and update the list of _Cases."""
        tl_status_esc = re.escape(self.tl_status)
        split = re.split("(^==\s*[^=]+?\s*==$)", text, flags=re.M|re.U)
        for i in xrange(len(split)):
            if i + 1 == len(split):
                break
            if not split[i].startswith("=="):
                continue
            title = split[i][2:-2].strip()
            body = old = split[i + 1]
            if not re.search("\s*\{\{" + tl_status_esc, body, re.U):
                continue
            status = self.read_status(body)
            re_id = "<!-- Bot Case ID \(please don't modify me\): (.*?) -->"
            try:
                id_ = re.search(re_id, body).group(1)
                case = [case for case in cases if case.id == id_][0]
            except (AttributeError, IndexError):
                id_ = self.select_next_id(conn)
                re_id2 = "(\{\{" + tl_status_esc
                re_id2 += "(.*?)\}\})(<!-- Bot Case ID \(please don't modify\): .*? -->)?"
                repl = ur"\1 <!-- Bot Case ID (please don't modify): {0} -->"
                body = re.sub(re_id2, repl.format(id_), body)
                case = _Case(id_, title, status)
                cases.append(case)
            else:
                if case.title != title:
                    self.update_case_title(conn, id_, title)
                    case.title = title
            case.body, case.old = body, old
            case.apparent_status = status

    def select_next_id(self, conn):
        """Return the next incremental ID for a case."""
        query = "SELECT MAX(case_id) FROM case"
        with conn.cursor() as cursor:
            cursor.execute(query)
            current = cursor.fetchone()[0]
            if current:
                return current + 1
            return 1

    def read_status(self, body):
        """Parse the current status from a case body."""
        aliases = {
            self.STATUS_NEW: ("",),
            self.STATUS_OPEN: ("open", "active", "inprogress"),
            self.STATUS_STALE: ("stale",),
            self.STATUS_NEEDASSIST: ("needassist", "relist", "relisted"),
            self.STATUS_REVIEW: ("review",),
            self.STATUS_RESOLVED: ("resolved", "resolve"),
            self.STATUS_CLOSED: ("closed", "close"),
        }
        templ = re.escape(self.tl_status)
        status = re.search("\{\{" + templ + "\|?(.*?)\}\}", body, re.S|re.U)
        if not status:
            return self.STATUS_UNKNOWN
        for option, names in aliases.iteritems():
            if status.group(1).lower() in names:
                return option
        return self.STATUS_UNKNOWN

    def update_case_title(self, conn, id_, title):
        """Update a case title in the database."""
        query = "UPDATE case SET case_title = ? WHERE case_id = ?"
        with conn.cursor() as cursor:
            cursor.execute(query, (title, id_))

    def clerk(self, conn, cases):
        """Actually go through cases and modify those to be updated."""
        query = "SELECT volunteer_username FROM volunteer"
        with conn.cursor() as cursor:
            cursor.execute(query)
            volunteers = [name for (name,) in cursor.fetchall()]
        notices = []
        for case in cases:
            notices += self.clerk_case(conn, case, volunteers)
        return notices

    def clerk_case(self, conn, case, volunteers):
        """Clerk a particular case and return a list of any notices to send."""
        if case.apparent_status == self.STATUS_NEW:
            # NOTIFY ALL PARTIES
            # SET STATUS TO OPEN IF VOLUNTEER EDITS
        elif case.apparent_status == self.STATUS_OPEN:
            # OPEN FOR OVER FOUR DAYS: SET STATUS TO REVIEW
                # SEND MESSAGE TO WT:DRN
            # ELSE OVER 10K TEXT SINCE LAST VOLUNTEER EDIT: SET STATUS TO NEEDASSIST
                # SEND MESSAGE TO WT:DRN
            # ELSE NO EDITS IN ONE DAY: SET STATUS TO STALE
                # SEND MESSAGE TO WT:DRN
        elif case.apparent_status == self.STATUS_NEEDASSIST:
            # OPEN FOR OVER FOUR DAYS: SET STATUS TO REVIEW
            # ELSE VOLUNTEER EDIT SINCE STATUS SET: SET STATUS OPEN
        elif case.apparent_status == self.STATUS_STALE:
            # OPEN FOR OVER FOUR DAYS: SET STATUS TO REVIEW
            # ELSE EDITS SINCE STATUS SET: SET STATUS OPEN
        elif case.apparent_status == self.STATUS_REVIEW:
            # OPEN FOR OVER SEVEN DAYS: SEND MESSAGE TO ZHANG
        elif case.apparent_status in [self.STATUS_RESOLVED, self.STATUS_CLOSED]:
            # ONE DAY SINCE STATUS SET: SET STATUS ARCHIVED
                # ADD ARCHIVE TEMPLATE, REMOVE NOARCHIVE

    def save(self, page, cases, kwargs):
        """Save any changes to the noticeboard."""
        newtext = text = page.get()
        counter = 0
        for case in cases:
            if case.old != case.body:
                newtext = newtext.replace(case.old, case.body)
                counter += 1
        if newtext == text:
            self.logger.info(u"Nothing to edit on [[{0}]]".format(page.title))
            return

        worktime = time() - start
        if worktime < 60:
            sleep(60 - worktime)
        page.reload()
        if page.get() != text:
            log = "Someone has edited the page while we were working; restarting"
            self.logger.warn(log)
            return self.run(**kwargs)
        summary = self.summary.replace("$3", str(counter))
        page.edit(text, summary, minor=True, bot=True)
        self.logger.info(u"Saved page [[{0}]]".format(page.title))

    def send_notices(self, site, notices):
        """Send out any templated notices to users or pages."""
        if not notices:
            self.logger.info("No notices to send; finishing")
            return
        for notice in notices:
            target, template = notice.target, notice.template
            log = u"Notifying [[{0}]] with {1}".format(target, template)
            self.logger.info(log)
            page = site.get_page(target)
            try:
                text = page.get()
            except exceptions.PageNotFoundError:
                text = ""
            if notice.too_late in text:
                log = u"Skipping [[{0}]]; was already notified".format(target)
                self.logger.info(log)
            text += ("\n" if text else "") + template
            try:
                page.edit(text, summary, minor=False, bot=True)
            except exceptions.EditError as error:
                name, msg = type(error).name, error.message
                log = u"Couldn't leave notice on {0} because of {1}: {2}"
                self.logger.error(log.format(page.title, name, msg))

        self.logger.info("Done sending notices")


class _Case(object):
    """A object representing a dispute resolution case."""
    def __init__(self, id_, title, status):
        self.id = id_
        self.title = title
        self.status = status

        self.body = None
        self.old = None
        self.apparent_status = None


class _Notice(object):
    """An object representing a notice to be sent to a user or a page."""
    def __init__(self, target, template, too_late):
        self.target = target
        self.template = template
        self.too_late = too_late
